from urllib.parse import urlencode

from allauth.account.adapter import DefaultAccountAdapter
from allauth.socialaccount.models import SocialApp, SocialAccount
from django.conf import settings


class Auth0LogoutAdapter(DefaultAccountAdapter):
    def get_logout_redirect_url(self, request):
        if (
            request.user.is_authenticated
            and SocialAccount.objects.filter(
                provider="auth0", user=request.user
            ).exists()
        ):
            app = SocialApp.objects.get(provider="auth0")
            query = {"client_id": app.client_id}
            if settings.AUTH0_LOGOUT_RETURN_TO:
                query["returnTo"] = settings.AUTH0_LOGOUT_RETURN_TO
            query_str = urlencode(query)
            return f"https://{settings.AUTH0_LOGIN_DOMAIN}/v2/logout/?{query_str}"
        return super().get_logout_redirect_url(request)



# Generated by Django 3.2 on 2021-08-25 04:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0012_remove_lobby_ready"),
    ]

    operations = [
        migrations.AddField(
            model_name="gameexperience",
            name="multiplayer",
            field=models.BooleanField(
                blank=True,
                help_text="Whether multiple players play a single instance of the game. If unset, this become configurable at the run level.",
                null=True,
            ),
        ),
    ]



# Generated by Django 3.2.4 on 2021-07-07 09:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0004_rename_token_apitoken'),
    ]

    operations = [
        migrations.AddField(
            model_name='gameexperience',
            name='continuous',
            field=models.BooleanField(blank=True, help_text='Whether players can join after a run is in play. If unset, this become configurable at the run level.', null=True),
        ),
        migrations.AddField(
            model_name='run',
            name='continuous',
            field=models.BooleanField(default=False, help_text='Automatically add new players to new instances for runs in play.'),
        ),
    ]



# Generated by Django 2.2.22 on 2021-07-12 02:40

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0005_auto_20210712_0038"),
    ]

    operations = [
        migrations.AlterField(
            model_name="instance",
            name="run",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.SIMPL_RUN,
            ),
        ),
        migrations.AlterField(
            model_name="lobby",
            name="run",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.SIMPL_RUN
            ),
        ),
        migrations.AlterField(
            model_name="player",
            name="run",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.SIMPL_RUN
            ),
        ),
    ]



from __future__ import annotations

import enum
import uuid
from typing import TYPE_CHECKING, List, Optional, Tuple

import faker
from django.apps import apps
from django.db import models, transaction
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from simpl.conf import settings
from simpl.utils.models import DataMixin
from simpl.utils.name_faker import random_name

from . import get_character_model, get_instance_model, managers

try:
    from django.db.models import JSONField
except ImportError:
    from django.contrib.postgres.fields import JSONField

fake = faker.Faker()


class BaseGameExperience(models.Model):
    name = models.CharField(max_length=100)
    experience_id = models.UUIDField(db_index=True, blank=True, null=True)
    date_created = models.DateTimeField(editable=False, default=timezone.now)
    version = models.CharField(max_length=20, blank=True)
    multiplayer = models.BooleanField(
        blank=True,
        null=True,
        help_text="Whether multiple players play a single instance of the game. "
        "If unset, this become configurable at the run level.",
    )
    continuous = models.BooleanField(
        blank=True,
        null=True,
        help_text="Whether players can join after a run is in play. "
        "If unset, this become configurable at the run level.",
    )

    objects = managers.GameExperienceManager()
    _default_manager: managers.GameExperienceManager

    class Meta:
        abstract = True
        ordering = "-date_created"

    def __str__(self):
        if self.version:
            return f"{self.name} (v{self.version})"
        return self.name

    @staticmethod
    def new_instance_name():
        return " ".join(fake.words(3)).title()

    @cached_property
    def is_latest(self):
        if not self.experience_id:
            return True
        games = type(self)._default_manager.from_experience_id(self.experience_id)
        return self == games[-1] if games else True


class GameExperience(BaseGameExperience):
    """
    A specific Simpl experience.

    This is only required for projects that provide multiple different Simpl
    experiences.
    """

    class Meta:
        swappable = "SIMPL_GAME_EXPERIENCE"


class Class(models.Model):
    id = models.UUIDField(primary_key=True)
    name = models.CharField(max_length=200)

    def __str__(self):
        return self.name


class BaseRun(DataMixin, models.Model):
    class STATUS(enum.IntEnum):
        SETUP = 0
        PREPARE = 1
        PLAY = 2
        DEBRIEF = 3
        COMPLETE = 4

    STATUS.do_not_call_in_templates = True

    STATUSES = (
        (STATUS.SETUP.value, "Setup"),
        (STATUS.PREPARE.value, "Players Prepare"),
        (STATUS.PLAY.value, "Play"),
        (STATUS.DEBRIEF.value, "Debrief"),
        (STATUS.COMPLETE.value, "Complete"),
    )

    game: Optional[GameExperience] = models.ForeignKey(
        settings.SIMPL_GAME_EXPERIENCE, on_delete=models.CASCADE, blank=True, null=True
    )
    simpl_class: Optional[Class] = models.ForeignKey(
        Class, on_delete=models.SET_NULL, blank=True, null=True
    )
    name = models.CharField(
        max_length=100, help_text="Name of the run, for administration use."
    )
    public_name = models.CharField(
        max_length=100, help_text="Publically visible name of the game."
    )
    date_created = models.DateTimeField(editable=False, default=timezone.now)
    status = models.PositiveSmallIntegerField(choices=STATUSES, default=STATUS.SETUP)
    data: dict = JSONField(editable=False, default=dict, blank=True)
    managers = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        blank=True,
        related_name="simpl_run_set",
        related_query_name="simpl_run",
    )
    multiplayer = models.BooleanField(default=True)
    continuous = models.BooleanField(
        default=False,
        help_text="Automatically add new players to new instances for runs in play.",
    )
    date_continuous_end = models.DateTimeField(blank=True, null=True)

    date_prepare = None

    player_set: Player.PlayerRelatedManager
    lobby_set: Lobby.LobbyRelatedManager

    class Meta:
        abstract = True
        ordering = ("-date_created",)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.pk and self.game:
            if self.game.continuous is not None:
                self.continuous = self.game.continuous
            if self.game.multiplayer is not None:
                self.multiplayer = self.game.multiplayer
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse("simpl", kwargs={"pk": self.pk})

    @property
    def setting_up(self):
        return self.status == self.STATUS.SETUP

    @property
    def preparing(self):
        return self.status == self.STATUS.PREPARE

    @property
    def playing(self):
        return self.status == self.STATUS.PLAY

    @property
    def in_debrief(self):
        return self.status == self.STATUS.DEBRIEF

    @property
    def completed(self):
        return self.status == self.STATUS.COMPLETE

    def get_play_url(self):
        return ""

    @cached_property
    def instances(self):
        from simpl import get_instance_model

        Instance = get_instance_model()
        return Instance.objects.filter(run=self)

    @cached_property
    def ended(self):
        if self.completed:
            return True
        instances = self.instances.count()
        return (
            instances > 0
            and self.instances.filter(date_end__isnull=False).count() == instances
        )

    def prepare(self) -> List[BaseInstance]:
        """
        Create instances for players that are ready to begin.

        For multiplayer games, creates instances for lobbies that are ready. For
        single player games, creates instances for unlinked players to this
        run.
        """
        if self.date_prepare and self.date_prepare > timezone.now():
            return []
        instances = []
        with transaction.atomic():
            if self.multiplayer:
                lobby: Lobby
                for lobby in self.lobby_set.ready(include_linked=True):
                    instance, created = self.create_multiplayer_instance_from_lobby(
                        lobby
                    )
                    if created:
                        instances.append(instance)
            else:
                for player in self.player_set.ready():
                    instance = self.create_singleplayer_instance(player)
                    instances.append(instance)
        return instances

    def create_multiplayer_instance_from_lobby(self, lobby):
        assert lobby.run == self and lobby.ready
        instance, created = self.get_or_create_multiplayer_instance(lobby)
        player: BasePlayer
        for player in lobby.player_set.ready():
            self.add_player_to_instance(player, instance)
        return instance, created

    def add_player_to_instance(self, player: BasePlayer, instance: BaseInstance):
        player.character = get_character_model().objects.create(
            instance=instance,
            user=player.user,
        )
        player.save()

    def get_or_create_multiplayer_instance(
        self, lobby: Lobby, defaults: dict = None
    ) -> Tuple[BaseInstance, bool]:
        defaults = defaults or {}
        if not "name" in defaults:
            defaults["name"] = lobby.name
        defaults["game"] = self.game
        instance, created = get_instance_model()._default_manager.get_or_create(
            run=self, lobby=lobby, defaults=defaults
        )
        if created:
            lobby.instance = instance
            lobby.save()
        return instance, created

    def create_singleplayer_instance(self, player: BasePlayer) -> BaseInstance:
        instance = get_instance_model()._default_manager.create(
            run=self, game=self.game
        )
        self.add_player_to_instance(player, instance)
        return instance

    def start_instances(self, instances: List[BaseInstance]):
        now = timezone.now()
        started = 0
        for instance in instances:
            if not instance.date_start:
                instance.date_start = now
                instance.save()
                started += 1
        return started

    def _get_app_setting(self, key):
        app_setting = getattr(apps.get_app_config("simpl"), key)
        if callable(app_setting):
            app_setting = app_setting(self)
        return app_setting

    @property
    def continuous_configurable(self) -> bool:
        if self.game:
            return self.game.continuous is None
        return self._get_app_setting("CONTINUOUS_CONFIGURABLE")

    @property
    def continuous_open(self) -> bool:
        """
        Whether the run is open for continuous runs.

        The run must be marked as continuous, in Play status, and without a
        date_continous_end in the past.
        """
        return (
            self.status == self.STATUS.PLAY
            and self.continuous
            and (
                not self.date_continuous_end
                or self.date_continuous_end > timezone.now()
            )
        )

    @property
    def use_status_prepare(self) -> bool:
        return self._get_app_setting("USE_STATUS_PREPARE")

    @property
    def use_status_debrief(self) -> bool:
        return self._get_app_setting("USE_STATUS_DEBRIEF")

    @property
    def can_end(self) -> bool:
        return self._get_app_setting("CAN_END_RUN")

    @property
    def can_restart(self) -> bool:
        return self.can_end and self._get_app_setting("CAN_RESTART_RUN")


class Run(BaseRun, DataMixin, models.Model):
    """
    One runthrough (of one or more game instances) of a Simpl experience.
    """

    objects: models.manager.Manager[Run]

    class Meta:
        swappable = "SIMPL_RUN"
        ordering = ("-date_created",)


class BaseInstance(DataMixin, models.Model):
    class STATUS(enum.IntEnum):
        WAITING = 1
        PREPARE = 2
        PLAY = 3
        DEBRIEF = 4
        COMPLETE = 5

    STATUS.do_not_call_in_templates = True

    @property
    def status(self):
        if self.run:
            if self.run.status == BaseRun.STATUS.SETUP:
                return self.STATUS.WAITING
            if self.run.status == BaseRun.STATUS.DEBRIEF:
                return self.STATUS.DEBRIEF
            if self.run.status == BaseRun.STATUS.PREPARE:
                return self.STATUS.PREPARE
        if self.date_end or (self.run and self.run.status == BaseRun.STATUS.COMPLETE):
            return self.STATUS.COMPLETE
        if self.date_start or (self.run and self.run.status == BaseRun.STATUS.PLAY):
            return self.STATUS.PLAY
        return self.STATUS.WAITING

    run: Optional[BaseRun] = models.ForeignKey(
        settings.SIMPL_RUN, on_delete=models.CASCADE, blank=True, null=True
    )
    game: Optional[GameExperience] = models.ForeignKey(
        settings.SIMPL_GAME_EXPERIENCE, on_delete=models.CASCADE, blank=True, null=True
    )
    date_created = models.DateTimeField(editable=False, default=timezone.now)
    date_start = models.DateTimeField(blank=True, null=True)
    date_end = models.DateTimeField(blank=True, null=True)
    name = models.CharField(max_length=100, blank=True)
    data: dict = JSONField(editable=False, default=dict, blank=True)

    class Meta:
        abstract = True
        ordering = ("-date_created",)

    def save(self, *args, **kwargs):
        if not self.name:
            if self.run and self.run.game:
                self.name = self.run.game.new_instance_name()
            elif self.game:
                self.name = self.game.new_instance_name()
            else:
                self.name = GameExperience.new_instance_name()
        super().save(*args, **kwargs)

    @property
    def running(self) -> bool:
        return bool(self.date_start and not self.date_end)

    def stop(self):
        self.date_end = timezone.now()
        self.save()

    def restart(self):
        self.date_end = None
        self.save()

    def get_status_display(self):
        if not self.date_start:
            return "Unstarted"
        if not self.date_end:
            return "Play"
        return "Complete"

    def archive(self):
        self.player_set.delete()

    def get_play_url(self):
        return ""


class Instance(BaseInstance):
    """
    An instance of a Simpl experience.
    """

    class Meta:
        swappable = "SIMPL_INSTANCE"


class BaseCharacterData(DataMixin, models.Model):
    name = models.CharField("In-game name", max_length=200)
    data: dict = JSONField(editable=False, default=dict, blank=True)

    class Meta:
        abstract = True
        ordering = ("name",)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if not self.name:
            self.name = self.get_default_name()
        super().save(*args, **kwargs)

    def get_default_name(self):
        other_characters = self._meta.model._default_manager.filter(
            instance=self.instance
        )
        if self.pk:
            other_characters = other_characters.exclude(pk=self.pk)
        return random_name(
            existing_names=other_characters.values_list("name", flat=True)
        )


class BaseCharacterLinked(models.Model):
    instance: Instance = models.ForeignKey(
        settings.SIMPL_INSTANCE, on_delete=models.CASCADE
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True


class Character(BaseCharacterData, BaseCharacterLinked):
    """
    A character in the game.

    Only characters with :attr:`user` set, are considered player characters.
    """

    class Meta:
        swappable = "SIMPL_CHARACTER"


class Lobby(models.Model):
    run: BaseRun = models.ForeignKey(settings.SIMPL_RUN, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)
    date_created = models.DateTimeField(editable=False, default=timezone.now)
    instance: Instance = models.ForeignKey(
        settings.SIMPL_INSTANCE, on_delete=models.SET_NULL, blank=True, null=True
    )

    objects = managers.LobbyQuerySet.as_manager()

    if TYPE_CHECKING:

        class LobbyRelatedManager(
            models.manager.RelatedManager["Lobby"],
            managers.LobbyQuerySet["Lobby"],
        ):
            ...

    player_set: Player.PlayerRelatedManager

    @cached_property
    def ready(self) -> Optional[bool]:
        if self.instance_id:
            return None  # Not ready because linked to an instance already.
        obj = Lobby.objects.prepare_ready().get(pk=self.pk)
        if obj.ready:
            return True
        if not obj.player_count:
            return None  # Not ready because there are no players.
        return False


class BasePlayer(models.Model):
    run: BaseRun = models.ForeignKey(settings.SIMPL_RUN, on_delete=models.CASCADE)
    ready = models.BooleanField(default=True)
    inactive = models.BooleanField(default=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    lobby: Optional[Lobby] = models.ForeignKey(
        Lobby, on_delete=models.SET_NULL, blank=True, null=True
    )
    character: Optional[Character] = models.ForeignKey(
        settings.SIMPL_CHARACTER, on_delete=models.SET_NULL, blank=True, null=True
    )
    completed = models.DateTimeField(blank=True, null=True)

    objects = managers.PlayerQuerySet.as_manager()

    if TYPE_CHECKING:

        class PlayerRelatedManager(
            models.manager.RelatedManager["Player"],
            managers.PlayerQuerySet["Player"],
        ):
            ...

    class Meta:
        abstract = True

    def __str__(self):
        if self.user:
            return self.user.get_full_name() or self.user.email or self.user.username
        if self.character:
            return str(self.character)
        return "<Player>"

    @property
    def public_name(self):
        if self.user:
            name = self.user.get_full_name()
        else:
            name = "Anonymous Player"
            if self.character:
                name += f": {self.character}"
        return name

    def get_play_url(self) -> str:
        custom_play_url = getattr(settings, "SIMPL_GET_PLAY_URL", None)
        if custom_play_url:
            return import_string(custom_play_url)(self)
        if self.character and self.character.instance:
            return self.character.instance.get_play_url()
        elif self.run:
            return self.run.get_play_url()
        return ""


class Player(BasePlayer):
    """
    A player linked to a run.
    """

    class Meta:
        swappable = "SIMPL_PLAYER"


class APIToken(models.Model):
    token = models.CharField(max_length=32, editable=False, unique=True)
    name = models.CharField(max_length=100, blank=True)
    last_used = models.DateField(blank=True, null=True)

    def __str__(self):
        return self.name or self.token_part

    @property
    def token_part(self):
        return f"{self.token[:8]}..."

    def save(self, *args, **kwargs):
        if not self.token:
            self.token = uuid.uuid4().hex
        super().save(*args, **kwargs)



from django.conf import settings as django_settings


class AppSettings:
    """
    A holder for app-specific default settings (project settings have
    priority).

    Settings split with two underscores are looked up from project settings
    as a dictionary, for example::

        # In myapp.conf:
        class Settings(AppSettings):
            MYAPP__TASTE = 'sour'
            MYAPP__SCENT = 'apple'

        # In settings:
        MYAPP = {
            'TASTE': 'sweet',
        }

    Individual attributes can be retrieved, or entire underscored
    dictionaries::

        from myapp.conf import settings
        print("Tastes {0}".format(settings.MYAPP__TASTE))
        myapp_settings = settings.MYAPP
        print("Smells like {0}".format(myapp_settings['SCENT']))
    """

    def __getattribute__(self, attr):
        # Return magic attributes without doing anything.
        if attr.startswith("__"):
            return super().__getattribute__(attr)

        # Retrieve (or build) any settings dictionaries (split by two
        # undescores).
        try:
            dicts = super().__getattribute__("_app_dicts")
        except AttributeError:
            dicts = []
            for key in dir(self):
                if not key.startswith("_") and "__" in key:
                    potential_dict = key.split("__", 1)[0]
                    if potential_dict.upper():
                        dicts.append(potential_dict)
            self._app_dicts = dicts

        # If we're trying to get a settings dictionary, build and return it.
        if attr in dicts:
            dict_prefix = attr + "__"
            settings_dict = getattr(django_settings, attr, {}).copy()
            for full_key in dir(self):
                if not full_key.startswith(dict_prefix):
                    continue
                key = full_key[len(dict_prefix) :]
                if key in settings_dict:
                    continue
                settings_dict[key] = super().__getattribute__(full_key)
            return settings_dict

        # If it's a dictionary attribute we're looking for, retrieve it.
        dict_attr = not attr.startswith("_") and "__" in attr and attr.split("__", 1)[0]
        if dict_attr:
            try:
                settings_dict = getattr(django_settings, dict_attr)
                dict_prefix = dict_attr and dict_attr + "__"
                return settings_dict[attr[len(dict_prefix) :]]
            except (AttributeError, KeyError):
                return super().__getattribute__(attr)

        # It must be just a standard attribute.
        try:
            # If it's not upper case then it's just an attribute of this class.
            if attr != attr.upper() and not dict_attr:
                raise AttributeError()
            return getattr(django_settings, attr)
        except AttributeError:
            return super().__getattribute__(attr)



from .app_settings import AppSettings


class Settings(AppSettings):
    SIMPL_CHARACTER = "simpl.Character"
    SIMPL_GAME_EXPERIENCE = "simpl.GameExperience"
    SIMPL_INSTANCE = "simpl.Instance"


settings = Settings()



# Generated by Django 3.2.5 on 2021-07-29 05:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0008_merge_20210727_1117'),
    ]

    operations = [
        migrations.AddField(
            model_name='player',
            name='inactive',
            field=models.BooleanField(default=False),
        ),
    ]



from django.contrib import admin
from django.utils.html import format_html

from . import models


@admin.register(models.Run)
class RunAdmin(admin.ModelAdmin):
    date_hierarchy = "date_created"
    list_display = ["manage", "name", "date_created", "status_display"]
    list_filter = ["game", "status"]

    def has_change_permission(self, *args, **kwargs):
        return False

    def manage(self, obj):
        """
        Return the management URL for this Run.
        """
        return format_html(
            '<a href="{url}" class="viewlink">Manage</a>',
            url=obj.get_absolute_url(),
        )

    def status_display(self, obj):
        return obj.get_status_display()

    status_display.short_description = "Status"


# @admin.register(models.Player)
# class PlayerAdmin(admin.ModelAdmin):
#     readonly_fields = ["user", "lobby", "character", "completed"]


@admin.register(models.APIToken)
class APITokenAdmin(admin.ModelAdmin):
    readonly_fields = ["token", "last_used"]
    list_display = ["__str__", "last_used"]
    date_hierarchy = "last_used"

    def has_change_permission(self, *args, **kwargs):
        return False



import graphene
import graphql

from .. import get_player_model, models
from . import query, types

Player = get_player_model()


class TeamInput(graphene.InputObjectType):
    id = graphene.ID()
    name = graphene.String(required=True)
    session = graphene.String()
    players = graphene.List(graphene.ID, required=True)


class BalanceTeams(query.BalancingMixin, graphene.Mutation):
    class Arguments:
        run_id = graphene.ID(required=True)
        teams = graphene.List(TeamInput)
        delete_teams = graphene.List(graphene.ID)

    Output = types.Balancing

    @classmethod
    def mutate(cls, root, info, run_id, teams=None, delete_teams=None):
        run = cls.get_run(info, run_id)
        if not run:
            raise graphql.GraphQLError(message="Permission denied")

        run_teams = dict((str(team.pk), team) for team in run.lobby_set.all())
        run_players = dict((str(player.pk), player) for player in run.player_set.all())

        if teams:
            for team in teams:
                if delete_teams and team.id in delete_teams:
                    continue
                if team.id:
                    run_team = run_teams.get(team.id)
                    if not run_team:
                        continue
                else:
                    run_team = models.Lobby(run=run)
                run_team.name = team.name
                run_team.save()
                run_team.player_set.set(
                    [run_players[pk] for pk in team.players if pk in run_players]
                )

        if delete_teams:
            run.lobby_set.filter(pk__in=delete_teams).delete()

        # Get the object (with prefetches) fresh from the DB.
        return cls.get_run(info, run_id)


class AlterPlayer(query.BalancingMixin, graphene.Mutation):
    class Arguments:
        player_id = graphene.ID(required=True)
        active = graphene.Boolean(required=True)

    Output = types.BalancingPlayer

    @classmethod
    def mutate(cls, root, info, player_id, active):
        user = cls.get_user(info)
        run = None
        if user:
            try:
                player = Player.objects.get(pk=player_id)
            except Player.DoesNotExist:
                pass
            else:
                run = cls.check_run(user, player.run)
        if not run:
            raise graphql.GraphQLError(message="Permission denied")
        if player.inactive == active:
            player.inactive = not active
            player.save()
        return player



# Generated by Django 2.2.22 on 2021-07-27 11:17

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0005_auto_20210707_0934"),
        ("simpl", "0007_shorten_lobby_name"),
    ]

    operations = []



# Generated by Django 3.2.4 on 2021-06-16 02:54

from django.db import migrations, models

try:
    from django.db.models import JSONField
except ImportError:
    from django.contrib.postgres.fields import JSONField


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="run",
            name="data",
            field=JSONField(blank=True, default=dict, editable=False),
        ),
    ]



from typing import ClassVar, Optional

from django.contrib import messages
from django.contrib.auth.mixins import UserPassesTestMixin, LoginRequiredMixin
from django.db.models import Q
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property
from django.views.generic import DetailView, RedirectView
from django.views.generic.detail import SingleObjectMixin

from simpl import conf, get_run_model, get_player_model, nav

from . import models


class AuthProviderLoginRequiredMixin(LoginRequiredMixin):
    def get_login_url(self):
        login_url = super().get_login_url()
        provider = self.request.GET.get("provider", "")
        return f"{login_url}?provider={provider}" if provider else login_url


class SimplMixin(
    UserPassesTestMixin, AuthProviderLoginRequiredMixin, SingleObjectMixin
):
    simpl_name: ClassVar[Optional[str]] = None

    def get_queryset(self):
        model = get_run_model()
        return model._default_manager.all()

    @cached_property
    def run(self) -> models.Run:
        return getattr(self, "object", self.get_object())

    @cached_property
    def nav_data(self):
        return nav.get_nav(self.run)

    def test_func(self):
        user = self.request.user
        return user.is_superuser or self.run.managers.filter(pk=user.pk).exists()

    def get_template_names(self):
        template_names = super().get_template_names()
        if self.simpl_name and not self.template_name:
            template_names.insert(0, f"simpl/run_{self.simpl_name}.html")
        return template_names

    def dispatch(self, request, *args, **kwargs):
        if self.simpl_name:
            if (
                self.simpl_name not in self.nav_data
                or self.nav_data[self.simpl_name].status == nav.NavStatus.DISABLED
            ):
                raise Http404()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs["simpl_current_nav"] = self.simpl_name
        kwargs["simpl_configuring"] = nav.is_configuring(self.run)
        kwargs["simpl_nav"] = self.nav_data
        kwargs["simpl_logout_url"] = reverse(
            getattr(conf.settings, "SIMPL_LOGOUT_URL_NAME", "logout")
        )
        for status in nav.NavStatus:
            kwargs[f"NAV_{status.name}"] = status
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        if self.simpl_name:
            return nav.get_next_url(self.run, self.simpl_name)
        return super().get_success_url()


class InitialView(SingleObjectMixin, RedirectView):
    def get_queryset(self):
        model = get_run_model()
        return model._default_manager.all()

    def get_redirect_url(self, *args, **kwargs):
        run = self.get_object()
        data = nav.get_nav(run)
        if not data:
            raise Http404()
        provider = self.request.GET.get("provider", "")
        url = list(data.values())[0].url
        return f"{url}?provider={provider}" if provider else url


class StatusView(SimplMixin, DetailView):
    simpl_name = "status"

    @cached_property
    def ready(self):
        return (
            self.run.multiplayer is False
            or not self.run.player_set.unbalanced().exists()
        )

    def get_context_data(self, **kwargs):
        kwargs["ready"] = self.ready
        kwargs["configured"] = True
        kwargs["next_item"] = nav.get_next_item(self.run, self.simpl_name)
        return super().get_context_data(**kwargs)


class ConfigView(SimplMixin, DetailView):
    simpl_name = "config"


class TeamView(SimplMixin, DetailView):
    simpl_name = "team"


class PlayersView(SimplMixin, DetailView):
    simpl_name = "players"

    def get_players_queryset(self):
        return self.run.player_set.select_related(
            "user", "character__instance"
        ).order_by("user__first_name", "user__last_name")

    def filter_players_queryset(self, players_queryset):
        search = self.request.GET.get("q")
        if search:
            players_queryset = players_queryset.filter(
                Q(user__first_name__icontains=search)
                | Q(user__last_name__icontains=search)
            )
        return players_queryset

    def get_active_players(self, players_queryset):
        players = players_queryset.active().has_user()
        if self.run.status >= self.run.STATUS.PLAY:
            players = players.exclude(character__instance=None)
        count = players.count()
        return count, self.filter_players_queryset(players)

    def get_inactive_players(self, players_queryset):
        players = players_queryset.inactive()
        return players.count(), self.filter_players_queryset(players)

    def get_invited_players(self, players_queryset):
        players = players_queryset.active().has_user(False)
        return players.count(), self.filter_players_queryset(players)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        players_qs = self.get_players_queryset()

        active_count, active_players = self.get_active_players(players_qs)
        if self.run.multiplayer:
            teams = {}
            for player in active_players:
                teams.setdefault(player.character.instance.name, [])
                teams[player.character.instance.name].append(player)
            context["teams"] = teams
        context["players"] = active_players
        context["players_count"] = active_count

        inactive_count, inactive_players = self.get_inactive_players(players_qs)
        context["inactive_players"] = inactive_players
        context["inactive_count"] = inactive_count

        invites_count, invited_players = self.get_invited_players(players_qs)
        context["invites"] = invited_players
        context["invites_count"] = invites_count

        if len(context.get("simpl_nav", {})) > 1 and context.get("simpl_configuring"):
            context["next_url"] = nav.get_next_url(self.run, self.simpl_name)
        return context

    def post(self, request, **kwargs):
        Player = get_player_model()
        deactivate = request.POST.getlist("deactivate")
        if deactivate:
            Player.objects.filter(pk__in=deactivate).update(inactive=True)
        reactivate = request.POST.getlist("reactivate")
        if reactivate:
            Player.objects.filter(pk__in=reactivate).update(inactive=False)
        return redirect(".")


class StartView(SimplMixin, DetailView):
    simpl_name = "start"

    @cached_property
    def ready(self):
        return (
            self.run.multiplayer is False
            or not self.run.player_set.unbalanced().exists()
        )

    def get_context_data(self, **kwargs):
        kwargs["ready"] = self.ready
        return super().get_context_data(**kwargs)

    def post(self, request, *args, **kwargs):
        run = self.run
        if self.ready and run.status in (run.STATUS.SETUP, run.STATUS.PREPARE):
            instances = run.prepare()
            run.start_instances(instances)
            run.status = run.STATUS.PLAY
            run.save()
        return redirect(self.get_success_url())


class DebriefView(SimplMixin, DetailView):
    template_name = "simpl/run_debrief.html"

    def post(self, request, *args, **kwargs):
        if self.run.status == self.run.STATUS.PLAY:
            self.run.status = self.run.STATUS.DEBRIEF
            message = "Player reports have been published."
        elif self.run.status == self.run.STATUS.DEBRIEF:
            self.run.status = self.run.STATUS.PLAY
            message = "Player reports have been hidden."
        self.run.save()
        messages.success(request, message)
        url = request.POST.get("redirect_to") or nav.get_next_url(
            self.run, self.simpl_name
        )
        return HttpResponseRedirect(url)

    def get_context_data(self, **kwargs):
        kwargs["redirect_to"] = self.request.GET.get("redirect_to")
        return super().get_context_data(**kwargs)


class EndGameplayView(SimplMixin, DetailView):
    template_name = "simpl/run_end_gameplay.html"
    simpl_name = "end"

    def post(self, request, *args, **kwargs):
        stoped_msg = "The game has been stopped for all players."
        started_msg = "The game has been restarted."
        if self.run.instances.exists():
            running_instances = self.run.instances.exclude(date_start=None).filter(
                date_end=None
            )
            if running_instances:
                for instance in running_instances:
                    instance.stop()
                message = stoped_msg
                if not self.run.continuous_open:
                    self.run.status = (
                        self.run.STATUS.DEBRIEF
                        if (
                            self.run.use_status_debrief
                            and self.run.status < self.run.STATUS.DEBRIEF
                        )
                        else self.run.STATUS.COMPLETE
                    )
            else:
                for instance in self.run.instances.exclude(date_end=None):
                    instance.restart()
                self.run.status = self.run.STATUS.PLAY
                message = started_msg
        else:
            if self.run.status == self.run.STATUS.COMPLETE:
                self.run.status = self.run.STATUS.PLAY
                message = started_msg
            else:
                self.run.status = self.run.STATUS.COMPLETE
                message = stoped_msg

        self.run.save()
        messages.success(request, message)
        url = request.POST.get("redirect_to") or nav.get_next_url(
            self.run, self.simpl_name
        )
        return HttpResponseRedirect(url)

    def get_context_data(self, **kwargs):
        kwargs["redirect_to"] = self.request.GET.get("redirect_to")
        return super().get_context_data(**kwargs)



import graphene
from allauth.socialaccount.models import SocialAccount
from django import http
from django.contrib.auth import get_user_model
from django.urls import reverse
from graphene_django import DjangoObjectType
from simpl import get_game_experience_model, get_instance_model, get_run_model, models

from . import utils

User = get_user_model()
Instance = get_instance_model()
GameExperience = get_game_experience_model()
Run = get_run_model()


def build_url(url: str, user_ids=None, request: http.HttpRequest = None):
    if url and user_ids:
        social = SocialAccount.objects.filter(user__in=user_ids).first()
        if social:
            if "?" in url:
                url, query = url.split("?", 1)[1]
            else:
                query = ""
            qd = http.QueryDict(query, mutable=True)
            qd["provider"] = social.provider if social else "auth0"
            url = f"{url}?{qd.urlencode()}"
    if request:
        url = request.build_absolute_uri(url)
    return url


class RunStatus(graphene.Enum):
    class Meta:
        enum = Run.STATUS


class InstanceStatus(graphene.Enum):
    class Meta:
        enum = models.BaseInstance.STATUS


class SimplInstance(DjangoObjectType):
    """
    A Simpl game instance
    """

    status = graphene.Field(InstanceStatus)
    players = graphene.List(
        graphene.ID, description="List of Auth0 IDs for players of this instance"
    )
    player_count = graphene.Int()

    class Meta:
        model = Instance
        skip_registry = True
        fields = ["name"]

    @staticmethod
    def resolve_status(obj, info):
        return obj.status

    @staticmethod
    def resolve_players(obj, info):
        users = User._default_manager.filter(character__instance=obj)
        return utils.get_auth0_ids(*users)

    @staticmethod
    def resolve_player_count(obj, info):
        return obj.character_set.exclude(user=None).count()


class SimplRun(DjangoObjectType):
    """
    A Simpl Run
    """

    management_url = graphene.String()
    players = graphene.List(
        graphene.ID, description="List of Auth0 IDs for players in this run"
    )
    players_unassigned = graphene.List(
        graphene.ID,
        description="List of Auth0 IDs for players that have not been assigned to a "
        "run yet.",
    )
    players_inactive = graphene.List(
        graphene.ID,
        description="List of Auth0 IDs for players that are inactive.",
    )
    player_count = graphene.Int()
    managers = graphene.List(
        graphene.ID, description="List of Auth0 IDs for managers of this run"
    )
    instances = graphene.List(SimplInstance)
    status = graphene.Field(RunStatus)
    game_id = graphene.UUID()
    class_id = graphene.ID()
    class_name = graphene.Field(graphene.String)
    continuous_open = graphene.Boolean()

    class Meta:
        model = Run
        skip_registry = True
        fields = ["id", "name", "multiplayer", "continuous"]

    @staticmethod
    def resolve_management_url(obj, info):
        url = reverse("simpl", kwargs={"pk": obj.id})
        manager_ids = obj.managers.values_list("id")
        return build_url(url, user_ids=manager_ids, request=info.context)

    @staticmethod
    def resolve_players(obj, info):
        users = User.objects.filter(player__run=obj)
        return utils.get_auth0_ids(*users)

    @staticmethod
    def resolve_players_unassigned(obj, info):
        users = User.objects.filter(player__run=obj, player__character=None)
        return utils.get_auth0_ids(*users)

    @staticmethod
    def resolve_players_inactive(obj, info):
        users = obj.player_set.filter(inactive=True).values_list("user")
        return utils.get_auth0_ids(*users)

    @staticmethod
    def resolve_player_count(obj, info):
        return obj.player_set.exclude(user=None).count()

    @staticmethod
    def resolve_managers(obj, info):
        return utils.get_auth0_ids(*obj.managers.all())

    @staticmethod
    def resolve_instances(obj, info):
        return Instance._default_manager.filter(run=obj)

    @staticmethod
    def resolve_game_id(obj, info):
        if obj.game:
            return obj.game.experience_id

    @staticmethod
    def resolve_class_id(obj, info):
        if obj.simpl_class:
            return obj.simpl_class.id

    @staticmethod
    def resolve_class_name(obj, info):
        return obj.simpl_class.name if obj.simpl_class else ""


class SimplUserInstance(graphene.ObjectType):
    """
    A Simpl game instance, with information specific to a user
    """

    # Actually based on a Character model instance.
    name = graphene.String()
    status = graphene.Field(InstanceStatus)
    url = graphene.Field(graphene.String, deprecation_reason="Use SimplUserRun.url")
    player_name = graphene.String()
    player_complete = graphene.Boolean()

    @staticmethod
    def resolve_name(obj, info):
        return obj.instance.name

    @staticmethod
    def resolve_status(obj, info):
        obj.instance.status

    @staticmethod
    def resolve_player_name(obj, info):
        return obj.name

    @staticmethod
    def resolve_player_status(obj, info):
        if obj.instance.status >= models.BaseInstance.STATUS.DEBRIEF:
            return True
        return bool(obj.data.get("complete"))


class SimplUserRun(graphene.ObjectType):
    """
    A Simpl Run, with information specific to a user
    """

    # Actually based on a Player model instance.
    id = graphene.ID()
    name = graphene.String()
    status = graphene.Field(RunStatus)
    instance = graphene.Field(SimplUserInstance)
    url = graphene.String()

    @staticmethod
    def resolve_id(obj, info):
        return obj.run.pk

    @staticmethod
    def resolve_name(obj, info):
        return obj.run.name

    @staticmethod
    def resolve_status(obj, info):
        return obj.run.status

    @staticmethod
    def resolve_instance(obj, info):
        return obj.character

    @staticmethod
    def resolve_url(obj: models.Player, info):
        url = obj.get_play_url()
        if url:
            return build_url(url, user_ids=[obj.user_id], request=info.context)


class SimplUser(graphene.ObjectType):
    """
    A Simpl User
    """

    auth0_id = graphene.ID()
    first_name = graphene.String()
    last_name = graphene.String()
    runs = graphene.List(SimplUserRun)

    @staticmethod
    def resolve_auth0_id(obj, info):
        auth0_id = getattr(obj, "auth0_id", None)
        if auth0_id:
            return auth0_id
        try:
            return utils.get_auth0_ids(obj)[0]
        except IndexError:
            return None

    @staticmethod
    def resolve_runs(obj, info):
        return obj.player_set.select_related("run", "character__instance")


class SimplGame(DjangoObjectType):
    """
    A Simpl game experience
    """

    id = graphene.UUID()
    runs = graphene.List(SimplRun)

    class Meta:
        model = GameExperience
        skip_registry = True
        fields = ["name", "continuous"]

    @staticmethod
    def resolve_runs(obj, info):
        return obj.run_set.all()

    @staticmethod
    def resolve_id(obj, info):
        return obj.experience_id


class SimplClass(DjangoObjectType):
    """
    A Simpl class
    """

    class Meta:
        model = models.Class
        skip_registry = True
        fields = ["id", "name"]



# Generated by Django 3.2.4 on 2021-07-01 11:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0003_auto_20210701_1047'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='Token',
            new_name='APIToken',
        ),
    ]



# Generated by Django 3.2 on 2021-08-19 22:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0009_player_inactive'),
    ]

    operations = [
        migrations.CreateModel(
            name='Class',
            fields=[
                ('id', models.UUIDField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=200)),
            ],
        ),
        migrations.AddField(
            model_name='run',
            name='simpl_class',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='simpl.class'),
        ),
    ]



import uuid
from allauth.socialaccount.models import SocialAccount

from django.apps import apps
from django.test import TestCase
from model_bakery import baker
from simpl import (
    get_game_experience_model,
    get_instance_model,
    get_player_model,
    get_run_model,
)
from simpl.models import Lobby, BaseRun
from django.contrib.auth import get_user_model


class RunTest(TestCase):
    def setUp(self):
        Player = get_player_model()
        self.simpl_run: BaseRun = baker.make(get_run_model(), game__name="Game")
        self.players = baker.make(
            Player, run=self.simpl_run, user__first_name="Julia", _quantity=3
        )
        self.player_inactive = baker.make(Player, run=self.simpl_run, inactive=True)

    def make_lobbies(self):
        lobbies = baker.make("simpl.Lobby", run=self.simpl_run, _quantity=2)
        lobbies[0].player_set.add(self.players[0], self.player_inactive)
        lobbies[1].player_set.add(self.players[1], self.players[2])
        baker.make("simpl.Lobby", run=self.simpl_run)
        return lobbies

    def test_prepare_multiplayer_run(self):
        lobbies = self.make_lobbies()

        instances = self.simpl_run.prepare()

        self.assertEqual(len(instances), 2)
        for index, instance in enumerate(instances):
            self.assertEqual(instance.game, self.simpl_run.game)
            self.assertEqual(instance.run, self.simpl_run)
            lobby = lobbies[index]
            self.assertEqual(instance.name, lobby.name)
            characters = instance.character_set.all()
            lobby_players = lobby.player_set.active()
            self.assertEqual(characters.count(), lobby_players.count())
            for player in lobby_players:
                self.assertIn(player.character, characters)
                self.assertEqual(player.character.user, player.user)

    def test_get_or_create_multiplayer_instance_existing(self):
        instance = baker.make(get_instance_model(), run=self.simpl_run)
        lobby = baker.make(Lobby, run=self.simpl_run, instance=instance)

        new_instance, created = self.simpl_run.get_or_create_multiplayer_instance(lobby)
        self.assertFalse(created)
        self.assertEqual(instance, new_instance)

    def test_prepare_single_player_run(self):
        self.simpl_run.multiplayer = False
        self.simpl_run.save()

        instances = self.simpl_run.prepare()

        self.assertEqual(len(instances), 3)
        for index, instance in enumerate(instances):
            self.assertEqual(instance.game, self.simpl_run.game)
            self.assertEqual(instance.run, self.simpl_run)
            characters = instance.character_set.all()
            self.assertEqual(characters.count(), 1)
            character = characters.first()
            player = self.simpl_run.player_set.get(character=character)
            self.assertEqual(character.user, player.user)

    def test_multiplayer_prepare(self):
        lobbies = self.make_lobbies()

        instances = self.simpl_run.prepare()

        self.assertEqual(len(instances), 2)
        for index, instance in enumerate(instances):
            self.assertEqual(instance.game, self.simpl_run.game)
            self.assertEqual(instance.run, self.simpl_run)
            lobby = lobbies[index]
            self.assertEqual(instance.name, lobby.name)
            characters = instance.character_set.all()
            lobby_players = lobby.player_set.active()
            self.assertEqual(characters.count(), lobby_players.count())
            for player in lobby_players:
                self.assertIn(player.character, characters)
                self.assertEqual(player.character.user, player.user)

    def test_single_player_prepare(self):
        self.simpl_run.multiplayer = False
        self.simpl_run.save()

        instances = self.simpl_run.prepare()

        self.assertEqual(len(instances), 3)
        for index, instance in enumerate(instances):
            self.assertEqual(instance.game, self.simpl_run.game)
            self.assertEqual(instance.run, self.simpl_run)
            characters = instance.character_set.all()
            self.assertEqual(characters.count(), 1)
            character = characters.first()
            player = self.simpl_run.player_set.get(character=character)
            self.assertEqual(character.user, player.user)

    def test_continuous_configurable(self):
        non_game_run = baker.make(get_run_model())
        self.assertEqual(
            non_game_run.continuous_configurable,
            apps.get_app_config("simpl").CONTINUOUS_CONFIGURABLE,
        )

        game_run_with_continuous_set = baker.make(
            get_run_model(), game__continuous=False
        )
        self.assertEqual(game_run_with_continuous_set.continuous_configurable, False)

        game_run_without_continuous_set = baker.make(
            get_run_model(), game__continuous=None
        )
        self.assertEqual(game_run_without_continuous_set.continuous_configurable, True)

    def test_continuous(self):
        run = baker.make(get_run_model(), continuous=True)
        player = baker.make(get_player_model(), run=run, user__username="p1")
        self.assertEqual(run.instance_set.count(), 0)

        run.status = run.STATUS.PLAY
        run.save()
        player = baker.make(get_player_model(), run=run, user__username="p2")
        self.assertEqual(run.instance_set.count(), 1)
        instance = run.instance_set.get()
        self.assertEqual(player.character.instance, instance)
        self.assertEqual(instance.status, instance.STATUS.PLAY)

    def test_version_sorting(self):
        GameExperience = get_game_experience_model()
        grouped_id = uuid.uuid4()
        for version in ["1.9", "1.10-copy", "1.10", "1.10a1", ""]:
            baker.make(GameExperience, experience_id=grouped_id, version=version)
        stand_alone_game = baker.make(
            GameExperience, experience_id=uuid.uuid4(), version="1.5"
        )
        non_experience_id_game = baker.make(
            GameExperience, experience_id=None, version="2"
        )

        games = GameExperience._default_manager.from_experience_id(grouped_id)
        self.assertEqual(
            [game.version for game in games], ["1.10a1", "", "1.10-copy", "1.9", "1.10"]
        )

        self.assertFalse(games[0].is_latest)
        self.assertTrue(games[-1].is_latest)

        self.assertTrue(stand_alone_game.is_latest)
        self.assertTrue(non_experience_id_game.is_latest)


class LobbyTestCase(TestCase):
    def test_ready(self):
        Player = get_player_model()
        baker.make(Lobby, name="empty")
        lobby2 = baker.make(Lobby, name="mixed")
        baker.make(Player, lobby=lobby2, ready=True, _quantity=2)
        baker.make(Player, lobby=lobby2, ready=False)
        lobby3 = baker.make(Lobby, name="ready")
        baker.make(Player, lobby=lobby3, ready=True, _quantity=2)
        lobby4 = baker.make(Lobby, name="unready")
        baker.make(Player, lobby=lobby4, ready=False, _quantity=2)
        self.assertEqual(set(Lobby.objects.ready()), {lobby3})

    def test_ready_queries(self):
        lobby = baker.make(Lobby)
        self.assertNumQueries(1)
        self.assertFalse(lobby.ready)
        self.assertNumQueries(2)

        annotated_lobby = Lobby.objects.prepare_ready().get()
        self.assertNumQueries(3)
        self.assertFalse(annotated_lobby.ready)
        self.assertNumQueries(3)


class SocialAccountTestCase(TestCase):
    def test_updates_user(self):
        user = baker.make(get_user_model())
        SocialAccount.objects.create(
            extra_data={"first_name": "Bob", "family_name": "Jones"}, user=user
        )
        user.refresh_from_db()
        self.assertEqual(user.first_name, "Bob")
        self.assertEqual(user.last_name, "Jones")

    def test_updates_user(self):
        user = baker.make(get_user_model())
        SocialAccount.objects.create(extra_data={}, user=user)
        user_refreshed = get_user_model().objects.get()
        self.assertEqual(user.first_name, user_refreshed.first_name)
        self.assertEqual(user.last_name, user_refreshed.last_name)

    def test_sets_user_email_if_empty(self):
        user = baker.make(get_user_model(), email="")
        SocialAccount.objects.create(extra_data={"email": "new@example.com"}, user=user)
        user.refresh_from_db()
        self.assertEqual(user.email, "new@example.com")

    def test_dont_update_user_email(self):
        user = baker.make(get_user_model(), email="old@example.com")
        SocialAccount.objects.create(extra_data={"email": "new@example.com"}, user=user)
        user_refreshed = get_user_model().objects.get()
        self.assertEqual(user.email, user_refreshed.email)



# Generated by Django 3.2 on 2021-08-24 18:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0011_gameexperience_experience_id'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='lobby',
            name='ready',
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Union


if TYPE_CHECKING:
    from .models import BasePlayer
    from allauth.socialaccount.models import SocialAccount


UNSET = object()


def start_continuous_singleplayer(instance: BasePlayer, **kwargs):
    if (
        not instance.character
        and instance.user
        and not instance.inactive
        and instance.ready
        and instance.run.continuous_open
    ):
        simpl_instance = instance.run.create_singleplayer_instance(instance)
        instance.run.start_instances([simpl_instance])


def update_user_from_socialaccount(instance: SocialAccount, **kwargs):
    data = instance.extra_data
    if not instance.user or not instance.extra_data:
        return

    def _update_user(
        key: str, data_key: Union[str, List[str]] = None, *, only_if_empty: bool = False
    ):
        user_value = getattr(instance.user, key)
        if only_if_empty and user_value:
            return False
        value = None
        if data_key is None:
            data_keys = [key]
        elif isinstance(data_key, str):
            data_keys = [data_key]
        else:
            data_keys = data_key
        for data_key in data_keys:
            value = data.get(data_key, UNSET)
            if value:
                break
        if value is UNSET or value == user_value:
            return False
        setattr(instance.user, key, value)
        return True

    changed = False
    changed |= _update_user("email", only_if_empty=True)
    changed |= _update_user("first_name", ["given_name", "first_name"])
    changed |= _update_user("last_name", ["family_name", "last_name"])

    if changed:
        instance.user.save()



from collections import abc

from django import template
from django.template import TemplateSyntaxError
from django.template.base import TokenType
from django.template.loader_tags import do_include
from django.utils.safestring import mark_safe

register = template.Library()


@register.tag
def includecontents(parser, token):
    """
    An extension of the ``{% include %}`` tag which loads a template and
    renders it with the current context. It accepts all the standard
    include tag arguments.

    The contents of this block tag is made available as a ``contents``
    variable in the context of the included template.

    Use ``{% contents some_name %}`` to use named areas which are
    available as ``contents.some_name`` in the context of the included
    template.

    Example::

        {% includecontents "includes/dialog.html" with dismissable=False %}
        Default contents lives outside a named area.
        {% contents title %}
            The contents of the named area called &quot;title&quot;.
        {% endcontents %}
        Default contents can live both before and after named areas at the same time.
        {% endincludecontents %}

    Included Template:

        <div class="dialog"{% if dismissable %} data-dismissable{% endif %}>
            {% if contents.title %}
                <h3>{{ contents.title }}</h3>
            {% endif %}
            <section>
            {{ contents }}
            </section>
            <button>{{ contents.save|default:"Save" %}</button>
            {% if "note" in contents %}
                {% if contents.note %}
                    <small>{{ contents.note }}</small>
                {% endif %}
            {% else %}
                <small>Default note only shown if named area wasn't provided.</small>
            {% endif %}
        </div>
    """
    nodelist, named_nodelists = get_contents_nodelists(parser)
    include_node = do_include(parser, token)
    include_node.origin = parser.origin
    return IncludeContentsNode(
        include_node=include_node,
        nodelist=nodelist,
        named_nodelists=named_nodelists,
    )


class RenderedContents(abc.Mapping):
    def __init__(self, context, nodelist, named_nodelists=None):
        with context.push():
            self.rendered_contents = mark_safe(nodelist.render(context).strip())
        self.rendered_areas = {}
        for key, named_nodelist in named_nodelists.items():
            with context.push():
                rendered_area = named_nodelist.render(context)
                if not rendered_area.strip():
                    rendered_area = ""
            self.rendered_areas[key] = mark_safe(rendered_area)

    def __str__(self):
        return self.rendered_contents

    def __getitem__(self, key):
        return self.rendered_areas[key]

    def __iter__(self):
        return iter(self.rendered_areas)

    def __len__(self):
        return len(self.rendered_contents)


class IncludeContentsNode(template.Node):
    def __init__(self, include_node, nodelist, named_nodelists):
        self.include_node = include_node
        self.nodelist = nodelist
        self.named_nodelists = named_nodelists

    def render(self, context):
        with context.push():
            context["contents"] = RenderedContents(
                context, nodelist=self.nodelist, named_nodelists=self.named_nodelists
            )
            return self.include_node.render(context)


def get_contents_nodelists(parser):
    end_tag = "endincludecontents"
    named_nodelists = {}
    default = []

    depth = 0
    while parser.tokens:
        token = parser.next_token()
        if token.token_type != TokenType.BLOCK:
            default.append(token)
            continue
        bits = token.split_contents()
        tag_name = bits[0]

        if tag_name == "includecontents":
            depth += 1
        elif tag_name == "contents" and not depth:
            if len(bits) < 2:
                raise TemplateSyntaxError(
                    "Unnamed %r tag within includecontents" % tag_name
                )
            if len(bits) > 2:
                raise TemplateSyntaxError("Invalid %r tag format" % tag_name)
            area_name = bits[1]
            if area_name in named_nodelists:
                raise TemplateSyntaxError(
                    "Duplicate name for %r tag: %r" % (tag_name, area_name)
                )
            named_nodelists[area_name] = parser.parse(["end" + tag_name])
            parser.delete_first_token()
            continue
        elif tag_name == end_tag:
            if depth:
                depth -= 1
            else:
                default.append(token)
                for default_token in reversed(default):
                    parser.prepend_token(default_token)
                nodelist = parser.parse([end_tag])
                parser.delete_first_token()
                return nodelist, named_nodelists
        default.append(token)
    parser.unclosed_block_tag([end_tag])



import sys
from typing import List
if sys.version_info >= (3, 8):
    from typing import List, TypedDict
else:
    from typing_extensions import TypedDict
from auth0.v3.management import Users, UsersByEmail
from django.conf import settings


class UserData(TypedDict):
    user_id: str
    email: str
    email_verified: bool
    username: str
    phone_number: str
    phone_verified: bool
    created_at: str
    updated_at: str
    identities: List[dict]
    app_metadata: dict
    user_metadata: dict
    picture: str
    name: str
    nickname: str
    multifactor: List[str]
    last_ip: str
    last_login: str
    logins_count: int
    blocked: bool
    given_name: str
    family_name: str


def get_access_token() -> str:
    from auth0.v3.authentication import GetToken

    non_interactive_client_id = settings.AUTH0_M2M_CLIENT_ID
    non_interactive_client_secret = settings.AUTH0_M2M_INTERACTIVE_CLIENT_SECRET
    get_token = GetToken(settings.AUTH0_M2M_DOMAIN)
    token = get_token.client_credentials(
        non_interactive_client_id,
        non_interactive_client_secret,
        "https://{}/api/v2/".format(settings.AUTH0_M2M_DOMAIN),
    )
    return token["access_token"]


def get_user_data(user_id: str) -> UserData:
    user = Users(settings.AUTH0_M2M_DOMAIN, get_access_token())
    return user.get(user_id)


def lookup_by_email(email: str) -> List[UserData]:
    search = UsersByEmail(settings.AUTH0_M2M_DOMAIN, get_access_token())
    return search.search_users_by_email(email)



from datetime import datetime
from typing import Any, Optional

from django.db import transaction
from django.utils import timezone


class DataMixin:
    def update_data(
        self, *, _save: bool = True, _now: Optional[datetime] = None, **kwargs
    ) -> list:
        """
        Update variable values (provided as keyword arguments).
        """
        modified = []
        if not kwargs:
            return modified
        if not hasattr(self, "modified_data"):
            self.modified_data = {}
            self.original_data = self.data.copy()
        now = _now or timezone.now()
        for key, value in kwargs.items():
            if key in self.data and self.data[key] == value:
                continue
            self.data[key] = value
            if key in self.original_data and self.original_data[key] == value:
                try:
                    del self.modified_data[key]
                except KeyError:
                    pass
            else:
                self.modified_data[key] = now
            modified.append(key)
        if _save:
            return self.save_data()
        return modified

    update_data.alters_data = True

    def save_data(self, *, keep_modified: bool = False) -> list:
        """
        Saves any pending data, fetching the latest data from the database first
        to avoid concurrency issues.
        """
        modified_data = getattr(self, "modified_data", None)
        if not modified_data:
            return []
        with transaction.atomic():
            new_data = self.data
            self.data = (
                self.__class__._default_manager.select_for_update()
                .filter(pk=self.pk)
                .values_list("data", flat=True)[0]
            )
            update_fields = {"data"}
            if "modified" in self._meta.fields:
                update_fields.add("modified")
            for key, now in modified_data.items():
                value = new_data[key]
                old_value = self.data.get(key)
                if key in self.data and old_value == value:
                    continue
                self.data[key] = value
                altered_fields = self.save_data_item(
                    key=key, value=value, old_value=old_value, now=now
                )
                if altered_fields:
                    update_fields |= altered_fields
            self.save(update_fields=update_fields)
        if keep_modified:
            self.original_data = self.data.copy()
        else:
            del self.modified_data
            del self.original_data
        return list(modified_data)

    save_data.alters_data = True

    def save_data_item(
        self, key: str, value: Any, old_value: Any, now: datetime
    ) -> Optional[set]:
        """
        A hook triggered whenever a data item is about to be saved.

        Should return a set of any fields altered.
        """



from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from graphene_django.views import GraphQLView

from simpl.schema import schema

from . import views

urlpatterns = [
    path(
        "graphql/",
        csrf_exempt(GraphQLView.as_view(schema=schema, graphiql=True)),
        name="simpl-api",
    ),
    path("<int:pk>/", views.InitialView.as_view(), name="simpl"),
    path("<int:pk>/status/", views.StatusView.as_view(), name="simpl.status"),
    path("<int:pk>/configure/", views.ConfigView.as_view(), name="simpl.config"),
    path("<int:pk>/team/", views.TeamView.as_view(), name="simpl.team"),
    path("<int:pk>/players/", views.PlayersView.as_view(), name="simpl.players"),
    path("<int:pk>/start/", views.StartView.as_view(), name="simpl.start"),
    path("<int:pk>/debrief/", views.DebriefView.as_view(), name="simpl.debrief"),
    path("<int:pk>/end/", views.EndGameplayView.as_view(), name="simpl.end"),
]



# Generated by Django 2.2.22 on 2021-09-03 15:14

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0013_gameexperience_multiplayer"),
    ]

    operations = [
        migrations.AlterField(
            model_name="player",
            name="character",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.SIMPL_CHARACTER,
            ),
        ),
    ]



SECRET_KEY = "-"

INSTALLED_APPS = [
    "simpl.apps.SimplConfig",
    "django.contrib.sites",
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "allauth",
    "allauth.account",
    "allauth.socialaccount",
    "allauth.socialaccount.providers.auth0",
]

SIMPL_GAME_EXPERIENCE = "simpl.GameExperience"
SIMPL_RUN = "simpl.Run"
SIMPL_INSTANCE = "simpl.Instance"
SIMPL_CHARACTER = "simpl.Character"
SIMPL_PLAYER = "simpl.Player"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.postgresql", "NAME": "simpl-test"}
}

ROOT_URLCONF = "simpl.urls"

MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
]



# Generated by Django 3.2.4 on 2021-07-01 10:47

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('simpl', '0002_run_data'),
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.CharField(editable=False, max_length=32, unique=True)),
                ('last_used', models.DateField(blank=True, null=True)),
            ],
        ),
        migrations.RemoveField(
            model_name='gameexperience',
            name='marketplace_key',
        ),
        migrations.RemoveField(
            model_name='player',
            name='player_key',
        ),
        migrations.RemoveField(
            model_name='run',
            name='marketplace_closed',
        ),
        migrations.RemoveField(
            model_name='run',
            name='marketplace_key',
        ),
        migrations.RemoveField(
            model_name='run',
            name='marketplace_open',
        ),
        migrations.AddField(
            model_name='run',
            name='managers',
            field=models.ManyToManyField(blank=True, to=settings.AUTH_USER_MODEL),
        ),
    ]



import math
import re

import mmh3
from django import template

register = template.Library()


def rshift(val, n):
    return (val % 0x100000000) >> n


@register.filter
def playercolor(player):
    saturation = 41
    lightness = 24
    hash = mmh3.hash(str(player), 3)
    h = hash % 360
    s = rshift(hash, 11) % 100
    el = rshift(hash, 17) % 100
    s = math.floor((s / 100) * saturation + (100 - saturation) / 2)
    el = math.floor((el / 100) * lightness + (100 - lightness) / 2.5)
    return f"hsl({h},{s}%,{el}%)"


@register.filter
def playerinitials(player):
    initials = re.sub(r"(\w)[^ ]+", r"\1", str(player))
    initials = re.sub(r"[^\w]", "", initials)
    return f"{initials[0]}{initials[-1]}"



# Generated by Django 3.2 on 2021-09-16 17:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0017_run_date_continuous_end'),
    ]

    operations = [
        migrations.AddField(
            model_name='apitoken',
            name='name',
            field=models.CharField(blank=True, max_length=100),
        ),
    ]



from simpl import get_game_experience_model, get_run_model, models
from simpl.schema.external.utils import get_auth0_users, simpl_token_required
import graphene

from . import types

GameExperience = get_game_experience_model()
Run = get_run_model()

RUN_STATUSES = list(types.RunStatus._meta.enum.__members__)


class Query(graphene.ObjectType):
    user = graphene.Field(
        types.SimplUser,
        auth0_id=graphene.ID(required=True),
        description="Return the User for the given auth0 ID.\n\nDoes not create non-existant users.",
    )
    run = graphene.Field(
        types.SimplRun,
        id=graphene.ID(required=True),
        description="Return the Run for a given run ID.",
    )
    runs = graphene.List(
        types.SimplRun,
        statuses=graphene.List(types.RunStatus),
        game_id=graphene.UUID(),
        description=f"Return all Runs. Can be filtered by game ID and/or runs in specific statuses.\n\nValid statuses are: {', '.join(RUN_STATUSES)}",
    )
    games = graphene.List(
        types.SimplGame,
        description="Return Game Experiences.",
    )
    game = graphene.Field(
        types.SimplGame,
        id=graphene.UUID(required=True),
        description="Return the Game for a given ID.",
    )
    simpl_classes = graphene.List(
        types.SimplClass,
        description="Return Classes.",
    )
    simpl_class = graphene.Field(
        types.SimplClass,
        id=graphene.UUID(required=True),
        description="Return the Class for a given ID.",
    )

    @staticmethod
    @simpl_token_required
    def resolve_user(root, info, auth0_id):
        users = get_auth0_users(auth0_id, create_users=False)
        if not users:
            return None
        user = users[0]
        user.auth0_id = auth0_id
        return user

    @staticmethod
    @simpl_token_required
    def resolve_run(root, info, id):
        try:
            return Run._default_manager.get(pk=id)
        except Run.DoesNotExist:
            return None

    @staticmethod
    @simpl_token_required
    def resolve_runs(root, info, statuses=(), game_id=None):
        runs = Run._default_manager.all()
        if statuses:
            runs = runs.filter(status__in=statuses)
        if game_id:
            runs = runs.filter(game__experience_id=game_id)
        return runs

    @staticmethod
    @simpl_token_required
    def resolve_games(root, info):
        games = []
        current_experience_id = None
        current_group = []
        for game in (
            GameExperience._default_manager.exclude(experience_id=None)
            .order_by("experience_id")
            .iterator()
        ):
            if game.experience_id != current_experience_id:
                if current_group:
                    games.append(
                        GameExperience._default_manager.sort_games_by_version(
                            current_group
                        )[-1]
                    )
                    current_group = []
                current_experience_id = game.experience_id
            current_group.append(game)
        if current_group:
            games.append(
                GameExperience._default_manager.sort_games_by_version(current_group)[-1]
            )
        return games

    @staticmethod
    @simpl_token_required
    def resolve_game(root, info, id):
        games = GameExperience._default_manager.filter(experience_id=id)
        if games:
            return GameExperience._default_manager.sort_games_by_version(
                games
            )[-1]

    @staticmethod
    @simpl_token_required
    def resolve_simpl_classes(root, info):
        return models.Class.objects.all()

    @staticmethod
    @simpl_token_required
    def resolve_simpl_class(root, info, id):
        return models.Class.objects.filter(id=id).first()



from functools import wraps
from typing import TYPE_CHECKING, List

from allauth.socialaccount.models import SocialAccount
from django.contrib.auth import get_user_model
from django.utils import timezone
from graphql.execution.base import ResolveInfo
from simpl import models

from auth0.v3.exceptions import Auth0Error

from . import auth0

User = get_user_model()

if TYPE_CHECKING:

    class UserWithAuth0(User):
        auth0_id: str


def get_auth0_ids(*users: List[User]) -> List[str]:
    return list(
        SocialAccount.objects.filter(provider="auth0", user__in=users).values_list(
            "uid", flat=True
        )
    )


def only_auth0_users(users):
    """
    Filter the queryset to only those who are auth0 users.
    """
    auth0_users = SocialAccount.objects.filter(provider="auth0", user__in=users)
    return users.filter(user__in=auth0_users)


def _get_user_data(auth0_id: str):
    try:
        return auth0.get_user_data(auth0_id)
    except Auth0Error:
        raise ValueError("Auth0 ID is not valid.")


def get_auth0_users(
    *auth0_ids: List[str], create_users: bool = True, update_user_data: bool = True
):
    users: List[UserWithAuth0] = []
    for auth0_id in auth0_ids:
        try:
            social_acct = SocialAccount.objects.get(provider="auth0", uid=auth0_id)
            if update_user_data:
                user_data = _get_user_data(auth0_id)
                if social_acct.extra_data != user_data:
                    social_acct.extra_data = user_data
                    social_acct.save()
        except SocialAccount.DoesNotExist:
            user_data = _get_user_data(auth0_id)
            email = user_data["email"]
            try:
                user = User._default_manager.get(email=email)
            except User.DoesNotExist:
                if not create_users:
                    continue
                user = User._default_manager.create(username=email, email=email)
            social_acct = SocialAccount.objects.create(
                provider="auth0", uid=auth0_id, extra_data=user_data, user=user
            )
        user: UserWithAuth0 = social_acct.user
        user.auth0_id = auth0_id
        users.append(user)
    return users


def simpl_token_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        context = next(arg for arg in args if isinstance(arg, ResolveInfo)).context
        if not hasattr(context, "user") or not context.user.is_superuser:
            try:
                http_token = context.META["HTTP_AUTHORIZATION"]
                token = models.APIToken.objects.get(token=http_token)
            except (models.APIToken.DoesNotExist, KeyError):
                return None
            today = timezone.now().today()
            if token.last_used != today:
                token.last_used = today
                token.save()
        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

import dataclasses
import enum
from typing import NewType, Dict

from django.conf import settings
from django.urls import NoReverseMatch, reverse
from django.utils.module_loading import import_string

from simpl.models import Run


class NavStatus(enum.Enum):
    DISABLED = 0
    UNSTARTED = 1
    INCOMPLETE = 2
    READY = 3
    COMPLETE = 4


@dataclasses.dataclass(frozen=True)
class NavItem:
    name: str
    status: NavStatus = NavStatus.UNSTARTED
    hint: str = ""
    url: str = None
    next_url: str = None
    hidden: bool = False


NavType = NewType("NavType", Dict[str, NavItem])


def change_item(data: NavType, key: str, **kwargs):
    data[key] = dataclasses.replace(data[key], **kwargs)


def is_configuring(run: Run):
    return run.status in (Run.STATUS.SETUP, Run.STATUS.PREPARE)


def _add_urls(data: NavType, run: Run):
    data = dict(data)
    next_url = None
    for key, item in reversed(list(data.items())):
        try:
            url = reverse(f"simpl.{key}", kwargs={"pk": run.pk})
        except NoReverseMatch:
            url = None
        change_item(data, key, url=url, next_url=next_url)
        if item.status != NavStatus.DISABLED:
            next_url = url
    return data


def _get_and_order_nav_data(func_str: str, run: Run):
    data: NavType = dict(import_string(func_str)(run))
    # TODO: order
    return data


def get_setup(run: Run):
    data = _get_and_order_nav_data(settings.SIMPL_SETUP_NAV, run)
    return _add_urls(data, run)


def get_play(run: Run):
    data = _get_and_order_nav_data(settings.SIMPL_PLAY_NAV, run)
    return _add_urls(data, run)


def get_nav(run: Run):
    nav_func = get_setup if is_configuring(run) else get_play
    return nav_func(run)


def get_next_item(run: Run, current_name: str):
    available_items = dict(
        (key, item)
        for key, item in get_nav(run).items()
        if item.status != NavStatus.DISABLED and item.url
    )
    item_names = list(available_items)
    next_name = None
    try:
        current_index = item_names.index(current_name)
    except ValueError:
        # Fall back to the first available item.
        return list(available_items.values())[0]
    if current_index != -1:
        if is_configuring(run):
            # Get the next item
            try:
                next_name = item_names[current_index + 1]
                return available_items[next_name]
            except IndexError:
                pass
        return available_items[current_name]


def get_next_url(run: Run, current_name: str):
    return get_next_item(run, current_name).url



# Generated by Django 2.2.22 on 2021-06-02 10:40

import django.db.models.deletion
import django.utils.timezone
import simpl.utils.models
from django.conf import settings
from django.db import migrations, models

try:
    from django.db.models import JSONField
except ImportError:
    from django.contrib.postgres.fields import JSONField


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        migrations.swappable_dependency(settings.SIMPL_INSTANCE),
        migrations.swappable_dependency(settings.SIMPL_CHARACTER),
        migrations.swappable_dependency(settings.SIMPL_GAME_EXPERIENCE),
        migrations.swappable_dependency(settings.SIMPL_RUN),
        migrations.swappable_dependency(settings.SIMPL_PLAYER),
    ]

    operations = [
        migrations.CreateModel(
            name="Character",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200, verbose_name="In-game name")),
                (
                    "data",
                    JSONField(blank=True, default=dict, editable=False),
                ),
            ],
            options={"swappable": "SIMPL_CHARACTER"},
            bases=(simpl.utils.models.DataMixin, models.Model),
        ),
        migrations.CreateModel(
            name="GameExperience",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=100)),
                (
                    "date_created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("version", models.CharField(blank=True, max_length=20)),
                (
                    "marketplace_key",
                    models.CharField(db_index=True, max_length=36, null=True),
                ),
            ],
            options={"swappable": "SIMPL_GAME_EXPERIENCE"},
        ),
        migrations.CreateModel(
            name="Lobby",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                (
                    "date_created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("ready", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="Run",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Name of the run, for administration use.",
                        max_length=100,
                    ),
                ),
                (
                    "public_name",
                    models.CharField(
                        help_text="Publically visible name of the game.", max_length=100
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("multiplayer", models.BooleanField(default=True)),
                (
                    "status",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Setup"),
                            (1, "Players Prepare"),
                            (2, "Play"),
                            (3, "Debrief"),
                            (4, "Complete"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "marketplace_key",
                    models.CharField(blank=True, db_index=True, max_length=36),
                ),
                ("marketplace_open", models.BooleanField(blank=True, null=True)),
                ("marketplace_closed", models.DateTimeField(blank=True, null=True)),
                (
                    "game",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.SIMPL_GAME_EXPERIENCE,
                    ),
                ),
            ],
            options={"ordering": ("-date_created",), "swappable": "SIMPL_RUN"},
        ),
        migrations.CreateModel(
            name="Player",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "player_key",
                    models.CharField(blank=True, db_index=True, max_length=36),
                ),
                ("ready", models.BooleanField(default=True)),
                ("completed", models.DateTimeField(blank=True, null=True)),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.SIMPL_CHARACTER,
                    ),
                ),
                (
                    "lobby",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="simpl.Lobby",
                    ),
                ),
                (
                    "run",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.SIMPL_RUN,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"swappable": "SIMPL_PLAYER"},
        ),
        migrations.AddField(
            model_name="lobby",
            name="run",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.SIMPL_RUN
            ),
        ),
        migrations.CreateModel(
            name="Instance",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                ("date_start", models.DateTimeField(blank=True, null=True)),
                ("date_end", models.DateTimeField(blank=True, null=True)),
                ("name", models.CharField(blank=True, max_length=100)),
                (
                    "data",
                    JSONField(blank=True, default=dict, editable=False),
                ),
                (
                    "game",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.SIMPL_GAME_EXPERIENCE,
                    ),
                ),
                (
                    "run",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.SIMPL_RUN,
                    ),
                ),
            ],
            options={"swappable": "SIMPL_INSTANCE"},
            bases=(simpl.utils.models.DataMixin, models.Model),
        ),
        migrations.AddField(
            model_name="character",
            name="instance",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.SIMPL_INSTANCE
            ),
        ),
        migrations.AddField(
            model_name="character",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Generated by Django 3.2 on 2021-09-05 23:52

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0014_auto_20210903_1514'),
    ]

    operations = [
        migrations.AddField(
            model_name='lobby',
            name='instance',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.SIMPL_INSTANCE),
        ),
    ]



import json
from unittest import mock
import uuid

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from model_bakery import baker
from simpl import get_run_model
from simpl.models import APIToken, Class, Run
from simpl.schema import schema


class FakeContext:
    pass


class AuthTestCase(TestCase):
    def setUp(self):
        self.api_url = reverse("simpl-api")

    def test_anon(self):
        request = self.client.post(
            self.api_url, '{"query": "{runs{id}}"}', content_type="application/json"
        )
        self.assertEqual(request.status_code, 200)
        self.assertEqual(json.loads(request.content)["data"], {"runs": None})

    def test_auth(self):
        token = APIToken.objects.create()
        request = self.client.post(
            self.api_url,
            '{"query": "{runs{id}}"}',
            content_type="application/json",
            HTTP_AUTHORIZATION=token.token,
        )
        self.assertEqual(request.status_code, 200)
        self.assertEqual(json.loads(request.content)["data"], {"runs": []})


class PlayersMutationTestCase(TestCase):
    def test_add_player(self):
        run = baker.make(get_run_model())
        context = FakeContext()
        context.user = baker.make(get_user_model(), is_superuser=True)
        with mock.patch(
            "simpl.schema.external.utils._get_user_data",
        ) as data_mock:
            data_mock.return_value = {"email": "tester@example.com"}
            schema.execute(
                """mutation ($run: ID!, $players: [ID]!) {
                    players(runId: $run, auth0Ids: $players) { id }
                }""",
                variable_values={"run": run.pk, "players": ["abc"]},
                context_value=context,
            )
        self.assertEqual(run.player_set.count(), 1)
        player = run.player_set.get()
        self.assertFalse(player.inactive)

    def test_add_player_inactive(self):
        run = baker.make(get_run_model())
        context = FakeContext()
        context.user = baker.make(get_user_model(), is_superuser=True)
        with mock.patch(
            "simpl.schema.external.utils._get_user_data",
        ) as data_mock:
            data_mock.side_effect = lambda x: {
                "abc": {"email": "tester@example.com"},
                "xyz": {"email": "tester2@example.com"},
            }[x]
            schema.execute(
                """mutation ($run: ID!, $players: [ID]!, $inactivePlayers: [ID]) {
                    players(
                        runId: $run,
                        auth0Ids: $players,
                        auth0IdsInactive: $inactivePlayers
                    ) { id }
                }""",
                variable_values={
                    "run": run.pk,
                    "players": ["abc", "xyz"],
                    "inactivePlayers": ["xyz"],
                },
                context_value=context,
            )
        self.assertEqual(run.player_set.count(), 2)
        player = run.player_set.get(user__email="tester@example.com")
        player2 = run.player_set.get(user__email="tester2@example.com")
        self.assertFalse(player.inactive)
        self.assertTrue(player2.inactive)


class ClassMutationTestCase(TestCase):
    def test_add_class(self):
        class_id = str(uuid.uuid4())
        context = FakeContext()
        context.user = baker.make(get_user_model(), is_superuser=True)
        result = schema.execute(
            """mutation ($id: UUID!, $name: String!) {
                simplClass(id: $id, name: $name)
            }""",
            variable_values={"id": class_id, "name": "Test class"},
            context_value=context,
        )
        self.assertEqual(result.data["simplClass"], class_id)
        self.assertEqual(Class.objects.get().name, "Test class")

    def test_update_class(self):
        simpl_class = baker.make(Class)
        run = baker.make(Run, simpl_class=simpl_class)
        context = FakeContext()
        context.user = baker.make(get_user_model(), is_superuser=True)
        result = schema.execute(
            """mutation ($id: UUID!, $name: String!) {
                simplClass(id: $id, name: $name)
            }""",
            variable_values={"id": simpl_class.id.hex, "name": "New name"},
            context_value=context,
        )
        self.assertEqual(result.data["simplClass"], str(simpl_class.id))
        simpl_class.refresh_from_db()
        self.assertEqual(Class.objects.get().name, "New name")

        result = schema.execute(
            """
            query ($runId: ID!) {
                run(id: $runId) {
                    className
                }
            }""",
            variable_values={"runId": run.id},
            context_value=context,
        )
        self.assertEqual(result.data["run"]["className"], "New name")



# Generated by Django 3.2 on 2021-08-22 20:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0010_class'),
    ]

    operations = [
        migrations.AddField(
            model_name='gameexperience',
            name='experience_id',
            field=models.UUIDField(blank=True, db_index=True, null=True),
        ),
    ]



import random
import string

unisex_names = [
    "Addison",
    "Adrian",
    "Aiden",
    "Ainsley",
    "Alex",
    "Amari",
    "Andy",
    "Ari",
    "Ash",
    "Aspen",
    "Aubrey",
    "August",
    "Avery",
    "Bailey",
    "Bay",
    "Blaine",
    "Blake",
    "Bobbie",
    "Brett",
    "Brook",
    "Brooklyn",
    "Caelan",
    "Cameron",
    "Campbell",
    "Carroll",
    "Carson",
    "Casey",
    "Charlie",
    "Chris",
    "Clay",
    "Corey",
    "Dana",
    "Dakota",
    "Dale",
    "Dallas",
    "Daryl",
    "Delta",
    "Devin",
    "Dorian",
    "Drew",
    "Dylan",
    "Easton",
    "Eddie",
    "Eli",
    "Elliott",
    "Emerson",
    "Emery",
    "Finley",
    "Frances",
    "Frankie",
    "Gabriel",
    "Glenn",
    "Gray",
    "Harley",
    "Harper",
    "Hayden",
    "Hudson",
    "Hunter",
    "James",
    "Jamie",
    "Jayden",
    "Jean",
    "Jesse",
    "Jordan",
    "Jules",
    "Julian",
    "Kaden",
    "Kai",
    "Karter",
    "Kelly",
    "Kelsey",
    "Kendall",
    "Kennedy",
    "Kyle",
    "Lake",
    "Landry",
    "Lincoln",
    "Logan",
    "London",
    "Lou",
    "Mackenzie",
    "Mason",
    "Max",
    "Maxwell",
    "Monroe",
    "Morgan",
    "Parker",
    "Pat",
    "Peyton",
    "Phoenix",
    "Quinn",
    "Ray",
    "Reagan",
    "Reed",
    "Reese",
    "Remy",
    "Riley",
    "River",
    "Roan",
    "Rory",
    "Rowan",
    "Rudy",
    "Ryan",
    "Sage",
    "Sam",
    "Sawyer",
    "Shawn",
    "Sean",
    "Skylar",
    "Spencer",
    "Stevie",
    "Sydney",
    "Tanner",
    "Tatum",
    "Taylor",
    "Toby",
    "Tyler",
    "Val",
    "West",
    "Winter",
]
last_names = [
    "Adams",
    "Alexander",
    "Allen",
    "Anderson",
    "Bailey",
    "Baker",
    "Barnes",
    "Bell",
    "Bennett",
    "Brooks",
    "Brown",
    "Bryant",
    "Butler",
    "Campbell",
    "Carter",
    "Clark",
    "Coleman",
    "Collins",
    "Cook",
    "Cooper",
    "Cox",
    "Davis",
    "Diaz",
    "Edwards",
    "Evans",
    "Flores",
    "Foster",
    "Garcia",
    "Gonzales",
    "Gonzalez",
    "Gray",
    "Green",
    "Griffin",
    "Hall",
    "Harris",
    "Hayes",
    "Henderson",
    "Hernandez",
    "Hill",
    "Howard",
    "Hughes",
    "Jackson",
    "James",
    "Jenkins",
    "Johnson",
    "Jones",
    "Kelly",
    "King",
    "Lee",
    "Lewis",
    "Long",
    "Lopez",
    "Martin",
    "Martinez",
    "Miller",
    "Mitchell",
    "Moore",
    "Morgan",
    "Morris",
    "Murphy",
    "Nelson",
    "Parker",
    "Patterson",
    "Perez",
    "Perry",
    "Peterson",
    "Phillips",
    "Powell",
    "Price",
    "Ramirez",
    "Reed",
    "Richardson",
    "Rivera",
    "Roberts",
    "Robinson",
    "Rodriguez",
    "Rogers",
    "Ross",
    "Russell",
    "Sanchez",
    "Sanders",
    "Scott",
    "Simmons",
    "Smith",
    "Stewart",
    "Taylor",
    "Thomas",
    "Thompson",
    "Torres",
    "Turner",
    "Walker",
    "Ward",
    "Washington",
    "Watson",
    "White",
    "Williams",
    "Wilson",
    "Wood",
    "Wright",
    "Young",
]

metals = {
    "Brass",
    "Copper",
    "Gallium",
    "Iron",
    "Lead",
    "Nickel",
    "Pewter",
    "Steel",
    "Tin",
    "Zinc",
}


def random_name(existing_names=None):
    existing_names = existing_names or []
    existing_first_names, existing_last_names = [], []
    for name in existing_names:
        bits = name.split(" ")
        existing_first_names.append(bits[0])
        existing_last_names.append(bits[-1])

    first_name = None
    while first_name is None or first_name in existing_first_names:
        first_name = random.choice(unisex_names)
    parts = [first_name]
    if not random.randint(0, 5):
        parts.append(f"{random.choice(string.ascii_lowercase.upper())}.")
    last_name = None
    while last_name is None or last_name in existing_last_names:
        if random.randint(0, 10):
            last_name = random.choice(last_names)
        else:
            last_name = "-".join(random.sample(last_names, 2))
    parts.append(last_name)
    return " ".join(parts)


def metal_name(existing_names=None):
    existing_names = set(existing_names) if existing_names else set()
    available = metals - existing_names
    extra = 1
    while not available:
        extra += 1
        available += set(f"{n} {extra}" for n in metals) - existing_names
    return random.choice(list(available))


if __name__ == "__main__":
    print(random_name())



from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union
from django.apps import apps as django_apps
from django.core.exceptions import ImproperlyConfigured

from simpl.conf import settings

if TYPE_CHECKING:
    from simpl import models


def _get_model(key: str):
    model = getattr(settings, f"SIMPL_{key}")
    try:
        return django_apps.get_model(model, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            f"SIMPL_{key} must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            f"SIMPL_{key} refers to model '{model}' that has not been installed"
        )


def get_character_model() -> Type[
    Union[models.BaseCharacterData, models.BaseCharacterLinked]
]:
    return _get_model("CHARACTER")


def get_game_experience_model() -> Type[models.BaseGameExperience]:
    return _get_model("GAME_EXPERIENCE")


def get_instance_model() -> Type[models.BaseInstance]:
    return _get_model("INSTANCE")


def get_player_model() -> Type[models.BasePlayer]:
    return _get_model("PLAYER")


def get_run_model() -> Type[models.BaseRun]:
    return _get_model("RUN")



# Generated by Django 3.2 on 2021-09-05 23:52

from django.db import migrations


def link_lobby_instances(apps, schema_editor):
    Lobby = apps.get_model("simpl.Lobby")
    for lobby in Lobby.objects.all():
        player = (
            lobby.player_set.filter(character__instance__isnull=False)
            .select_related("character")
            .first()
        )
        if player:
            lobby.instance_id = player.character.instance_id
            lobby.save()


def unlink_lobby_instances(apps, schema_editor):
    Lobby = apps.get_model("simpl.Lobby")
    Lobby.objects.update(instance=None)


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0015_lobby_instance"),
    ]

    operations = [
        migrations.RunPython(link_lobby_instances, reverse_code=unlink_lobby_instances)
    ]



from typing import Sequence

import graphene
import graphql
from simpl import get_game_experience_model, get_run_model, models

from . import types
from .utils import get_auth0_users, simpl_token_required

GameExperience = get_game_experience_model()
Run = get_run_model()


class SimplRun(graphene.Mutation):
    """
    Create a new Simpl Run.
    """

    class Arguments:
        name = graphene.String(required=True)
        game_id = graphene.UUID()
        class_id = graphene.UUID()
        multiplayer = graphene.Boolean()
        continuous = graphene.Boolean()

    Output = types.SimplRun

    @staticmethod
    @simpl_token_required
    def mutate(
        root,
        info,
        name,
        game_id=None,
        class_id=None,
        multiplayer=False,
        continuous=False,
    ):
        if game_id:
            games = GameExperience._default_manager.from_experience_id(game_id)
            if not games:
                raise graphql.GraphQLError("No game found")
            game = games[-1]
        else:
            game = None
        simpl_class = models.Class.objects.get(id=class_id) if class_id else None
        run = Run.objects.create(
            name=name,
            game=game,
            simpl_class=simpl_class,
            multiplayer=multiplayer,
            continuous=continuous,
        )
        return run


class SimplClass(graphene.Mutation):
    """
    Create or update a Simpl Class.
    """

    class Arguments:
        id = graphene.UUID(required=True)
        name = graphene.String(required=True)

    Output = graphene.UUID

    @staticmethod
    @simpl_token_required
    def mutate(root, info, name, id):
        return models.Class.objects.update_or_create(id=id, defaults=dict(name=name))[
            0
        ].id


class SimplPlayers(graphene.Mutation):
    """
    Add Players to (or remove Players from) a Run.

    Can also mark players as in a Run, but inactive.

    Creates any users that don't already exist.
    """

    class Arguments:
        run_id = graphene.ID(required=True)
        auth0_ids = graphene.List(graphene.ID)
        auth0_ids_remove = graphene.List(graphene.ID)
        auth0_ids_inactive = graphene.List(graphene.ID)

    @staticmethod
    @simpl_token_required
    def mutate(
        root,
        info,
        run_id,
        auth0_ids: Sequence[str] = (),
        auth0_ids_remove: Sequence[str] = (),
        auth0_ids_inactive: Sequence[str] = (),
    ):
        run = Run.objects.get(pk=run_id)
        for user in get_auth0_users(*(set(auth0_ids) | set(auth0_ids_inactive))):
            inactive = user.auth0_id in auth0_ids_inactive
            player, created = run.player_set.get_or_create(
                user=user, defaults={"inactive": inactive}
            )
            if not created and player.inactive != inactive:
                started_game = (
                    player.character
                    and player.character.instance.status
                    >= models.BaseInstance.STATUS.PLAY
                )
                if not started_game:
                    player.inactive = inactive
                    player.save()

        remove_users = get_auth0_users(
            *auth0_ids_remove, create_users=False, update_user_data=False
        )
        if remove_users:
            run.player_set.filter(user__in=remove_users, character=None).delete()
            if not run.multiplayer:
                run.instance_set.filter(character__user__in=remove_users).delete()
        return run

    Output = types.SimplRun


class SimplManagers(graphene.Mutation):
    """
    Add or remove managers for a Run.
    """

    class Arguments:
        run_id = graphene.ID(required=True)
        auth0_ids = graphene.List(graphene.ID)
        auth0_ids_remove = graphene.List(graphene.ID)

    @staticmethod
    @simpl_token_required
    def mutate(root, info, run_id, auth0_ids=(), auth0_ids_remove=()):
        run = Run.objects.get(pk=run_id)
        add_users = get_auth0_users(*auth0_ids)
        if add_users:
            run.managers.add(*add_users)
        remove_users = get_auth0_users(
            *auth0_ids_remove, create_users=False, update_user_data=False
        )
        if remove_users:
            run.managers.remove(*remove_users)

    Output = types.SimplRun



# Generated by Django 3.2 on 2021-09-06 21:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpl', '0016_link_lobby_instances'),
    ]

    operations = [
        migrations.AddField(
            model_name='run',
            name='date_continuous_end',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



import graphene
from django.db.models import Prefetch

from simpl.models import Lobby
from .. import get_run_model, get_player_model
from . import types


Player = get_player_model()
Run = get_run_model()


class BalancingMixin:
    @staticmethod
    def get_user(info):
        user = getattr(info.context, "user", None)
        if user.is_authenticated:
            return user

    @classmethod
    def get_run(cls, info, run_id):
        user = cls.get_user(info)
        if not user:
            return None
        try:
            players_qs = Player.objects.select_related("user").order_by(
                "user__first_name", "user__last_name"
            )
            lobby_qs = Lobby.objects.prefetch_related(
                Prefetch("player_set", players_qs)
            ).prepare_ready().order_by("-date_created")
            run = Run.objects.prefetch_related(
                Prefetch("player_set", players_qs),
                Prefetch("lobby_set", lobby_qs)
            ).get(pk=run_id)
        except Run.DoesNotExist:
            return None
        return cls.check_run(user, run)

    @classmethod
    def check_run(cls, user, run):
        if not user.is_superuser and not run.managers.filter(pk=user.pk).exists():
            return None
        return run


class Query(BalancingMixin, graphene.ObjectType):
    balancing = graphene.Field(types.Balancing, run_id=graphene.ID())

    @classmethod
    def resolve_balancing(cls, root, info, run_id):
        return cls.get_run(info, run_id)



import graphene

from . import mutations, query
from .external import mutations as external_mutations
from .external.query import Query as ExternalQuery


class Mutation(graphene.ObjectType):
    balance_teams = mutations.BalanceTeams.Field()
    alter_player = mutations.AlterPlayer.Field()

    run = external_mutations.SimplRun.Field()
    players = external_mutations.SimplPlayers.Field()
    managers = external_mutations.SimplManagers.Field()
    simpl_class = external_mutations.SimplClass.Field()


class Query(query.Query, ExternalQuery):
    pass


schema = graphene.Schema(query=Query, mutation=Mutation)



import graphene
from graphene_django import DjangoObjectType

from .. import get_player_model, get_run_model, models

Player = get_player_model()
Run = get_run_model()


class BalancingPlayer(DjangoObjectType):
    name = graphene.String()
    email = graphene.String()
    session = graphene.String()

    class Meta:
        model = Player
        fields = ["id", "inactive", "ready"]

    @staticmethod
    def resolve_name(obj: Player, info):
        return obj.public_name

    @staticmethod
    def resolve_email(obj: Player, info):
        if obj.user and obj.user.email != str(obj):
            return obj.user.email


class BalancingTeam(DjangoObjectType):
    name = graphene.String()
    players = graphene.List(graphene.ID)
    session = graphene.String()
    ready = graphene.Boolean()

    class Meta:
        model = models.Lobby
        fields = ["id"]

    @staticmethod
    def resolve_name(obj: models.Lobby, info):
        return obj.name

    @staticmethod
    def resolve_ready(obj: models.Lobby, info):
        return obj.ready

    @staticmethod
    def resolve_players(obj: models.Lobby, info):
        return [player.pk for player in obj.player_set.all()]


class Balancing(DjangoObjectType):
    class Meta:
        model = Run
        fields = []

    sessions = graphene.List(graphene.String)
    players = graphene.List(BalancingPlayer)
    teams = graphene.List(BalancingTeam)

    @staticmethod
    def resolve_players(obj, info):
        return obj.player_set.all()

    @staticmethod
    def resolve_teams(obj, info):
        return obj.lobby_set.all()



# Generated by Django 2.2.22 on 2021-07-20 05:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0006_auto_20210712_0240"),
    ]

    operations = [
        migrations.AlterField(
            model_name="lobby",
            name="name",
            field=models.CharField(max_length=100),
        ),
    ]



from django.apps import AppConfig
from django.db.models.signals import post_save

from simpl import get_player_model, receivers


class SimplConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "simpl"
    CONTINUOUS_CONFIGURABLE = False
    USE_STATUS_PREPARE = False
    USE_STATUS_DEBRIEF = False
    CAN_END_RUN = True
    CAN_RESTART_RUN = True

    def ready(self):
        from allauth.socialaccount.models import SocialAccount

        post_save.connect(
            receivers.start_continuous_singleplayer, sender=get_player_model()
        )
        post_save.connect(
            receivers.update_user_from_socialaccount, sender=SocialAccount
        )



# Generated by Django 2.2.22 on 2021-07-12 00:38

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("simpl", "0004_rename_token_apitoken"),
    ]

    operations = [
        migrations.AlterField(
            model_name="run",
            name="managers",
            field=models.ManyToManyField(
                blank=True,
                related_name="simpl_run_set",
                related_query_name="simpl_run",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from __future__ import annotations

import uuid
from typing import Union

from django.db import models
from django.db.models import Q
from django.db.models.expressions import Case, Value, When
from packaging.version import parse


def _game_sort_key(game):
    version = parse(game.version)
    return not version.is_prerelease, version


class GameExperienceManager(models.Manager):
    def from_experience_id(self, experience_id: Union[uuid.UUID, str]):
        """
        Return a list of game experiences with the same experience_id, ordered
        from oldest to latest.
        """
        games = self.sort_games_by_version(
            self.filter(experience_id=experience_id).order_by("name")
        )
        for game in games:
            game.is_latest = game == games[-1]
        return games

    def sort_games_by_version(self, games):
        """
        Sorting considers valid pre-release versions to be oldest, then
        non-valid versions, then other valid versions. Here's an example list of
        experience versions sorted from oldest to latest:

        1. "1.10a1"
        2. ""
        3. "1.10-copy"
        4. "1.9
        5. "1.10"
        """
        games = list(games)
        games.sort(key=_game_sort_key)
        return games


class PlayerQuerySet(models.QuerySet):
    def unbalanced(self):
        return self.ready().filter(lobby=None)

    def active(self):
        return self.filter(inactive=False)

    def inactive(self):
        return self.filter(inactive=True)

    def ready(self):
        return (
            self.active()
            .has_user()
            .filter(character=None, ready=True)
            .select_related("user")
        )

    def has_user(self, value=True):
        return self.filter(user__isnull=not value)


class LobbyQuerySet(models.QuerySet):
    def empty(self, no_players: bool = True):
        qs = self.annotate(num_players=models.Count("player"))
        if no_players:
            return qs.filter(num_players=0)
        return qs.filter(num_players__gt=0)

    def availability(self, max_players: int = None, exclude_started=True):
        """
        Filter to only available lobbies.

        Pass a ``max_players`` integer to filter to lobbies with less players
        than this amount.

        Pass ``exclude_started=False`` if you don't want to exclude lobbies
        which already have a created instance.
        """
        qs = self
        if exclude_started:
            qs = self.alias(
                num_playing=models.Count("player", filter=~Q(player__character=None)),
            ).filter(num_playing=0)
        if max_players:
            qs = qs.annotate(num_available=max_players - models.Count("player")).filter(
                num_available__gt=0
            )
        return qs

    # Django <3.2 support
    def alias(self, *args, **kwargs):
        return getattr(super(), "alias", self.annotate)(*args, **kwargs)

    def prepare_ready(self):
        qs = self.annotate(player_count=models.Count("player"))
        qs = qs.alias(
            unready_players=models.Count(
                "player", filter=Q(player__character=None) & Q(player__ready=False)
            ),
        )
        return qs.annotate(
            ready=Case(
                When(player_count__gt=0, unready_players=0, then=Value(True)),
                default=Value(False),
                output_field=models.BooleanField(),
            ),
            not_ready=Case(
                When(unready_players__gt=0, then=Value(True)),
                default=Value(False),
                output_field=models.BooleanField(),
            ),
        )

    def ready(self, value: bool = True, include_linked: bool = False):
        """
        Filter to lobbies with at least one ready player, and no unready
        players.

        Checking for ``.ready(False)`` is not quite the inverse -- it will only
        include lobbies with at least one unready player.

        By default, both cases will only include lobbies not yet linked to an
        instance. Use ``include_linked=True`` to include all lobbies.
        """
        qs = self.filter(instance=None) if not include_linked else self.all()
        qs = qs.prepare_ready()
        if value:
            return qs.filter(ready=True)
        return qs.filter(not_ready=True)
